import pytest
from koyo.containers import MutableMapping, MutableSequence


def test_mapping():
    class T(MutableMapping[str, int]):
        """Test class."""

    t = T()
    t["a"] = 1
    t.update({"b": 2})
    assert t["a"] == 1
    assert t["b"] == 2
    del t["a"]
    assert "a" not in t


def test_sequence():
    class T(MutableSequence[int]):
        """Test class."""

        def _check(self, value: int) -> int:
            if value in self._list:
                raise ValueError("Duplicate value")
            return value

    t = T()
    t.append(1)
    with pytest.raises(ValueError):
        t.append(1)
    del t[0]
    assert len(t) == 0
    t.append(1)



"""Image processing functions."""
import typing as ty

import numpy as np


def clip_hotspots(img: np.ndarray, quantile: float = 0.99) -> np.ndarray:
    """Remove hotspots from ion image.

    Parameters
    ----------
    img : np.ndarray
        Image array.
    quantile : float
        Quantile at which to clip intensities.

    Returns
    -------
    img : np.ndarray
        Clipped array.

    """
    mask = np.isnan(img)
    img = np.nan_to_num(img)
    min_visible = np.max(img) / 256
    if min_visible > 0:
        hotspot_threshold = np.quantile(img[img > min_visible], quantile)
        img = np.clip(img, None, hotspot_threshold)
    img[mask] = np.nan
    return img


def reshape_array(array: np.ndarray, image_shape: ty.Tuple[int, int], pixel_index: np.ndarray, fill_value: float = 0):
    """
    Reshape 1D data to 2D heatmap.

    Parameters
    ----------
    array: np.array / list
        1D array of values to be reshaped
    image_shape: tuple
        final shape of the image
    pixel_index: np.array
        array containing positions where pixels should be placed, considering missing values -
        e.g. not acquired pixels
    fill_value : float, optional
        if value is provided, it will be used to fill-in the values

    Returns
    -------
    im_array: np.array
        reshaped heatmap of shape `image_shape`
    """
    if isinstance(array, np.matrix):
        array = np.asarray(array).flatten()
    array = np.asarray(array)
    dtype = np.float32 if isinstance(fill_value, float) else array.dtype

    image_n_pixels = np.prod(image_shape)
    im_array = np.full(image_n_pixels, dtype=dtype, fill_value=fill_value)
    im_array[pixel_index] = array
    im_array = np.reshape(im_array, image_shape)
    return im_array


def reshape_array_from_coordinates(
    array: np.ndarray, image_shape: ty.Tuple[int, int], coordinates: np.ndarray, fill_value: float = 0
):
    """Reshape array based on xy coordinates."""
    dtype = np.float32 if np.isnan(fill_value) else array.dtype
    im = np.full(image_shape, fill_value=fill_value, dtype=dtype)
    im[coordinates[:, 1] - 1, coordinates[:, 0] - 1] = array
    return im


def reshape_array_batch(
    array: np.ndarray, image_shape: ty.Tuple[int, int], pixel_index: np.ndarray, fill_value: float = 0
):
    """Reshape many images into a data cube."""
    array = np.asarray(array)
    if array.ndim == 1:
        return reshape_array(array, image_shape, pixel_index, fill_value)
    count = array.shape[1]
    dtype = np.float32 if isinstance(fill_value, float) else array.dtype

    im_array = np.full((count, np.prod(image_shape)), dtype=dtype, fill_value=fill_value)
    for i in range(count):
        im_array[i, pixel_index] = array[:, i]
    # reshape data
    im_array = np.reshape(im_array, (count, *image_shape))
    return im_array


def reshape_array_batch_from_coordinates(
    array: np.ndarray, image_shape: ty.Tuple[int, int], coordinates: np.ndarray, fill_value: int = 0
):
    """Batch reshape image."""
    if array.ndim != 2:
        raise ValueError("Expected 2-D array.")
    n = array.shape[1]
    dtype = np.float32 if np.isnan(fill_value) else array.dtype
    im = np.full((n, *image_shape), fill_value=fill_value, dtype=dtype)
    for i in range(n):
        im[i, coordinates[:, 1] - 1, coordinates[:, 0] - 1] = array[:, i]
    return im



"""Generic utilities."""
import re
import typing as ty
import unicodedata
from collections.abc import Iterable
from math import ceil

import numba as nb
import numpy as np

from koyo.typing import SimpleArrayLike


def is_between(value: float, lower: float, upper: float, inclusive: bool = True) -> bool:
    """Check if value is between lower and upper."""
    if inclusive:
        return lower <= value <= upper
    return lower < value < upper


def flatten_nested_list(list_of_lists: ty.List[ty.List]) -> ty.List:
    """Flatten nested list of lists."""
    return [item for sublist in list_of_lists for item in sublist]


def get_list_difference(li1: ty.List, li2: ty.List):
    """Get difference between two lists."""
    # get difference between two lists while keeping the order
    li_dif = [i for i in li1 + li2 if i not in li1 or i not in li2]
    return li_dif


def find_nearest_index(data: SimpleArrayLike, value: ty.Union[int, float, np.ndarray, Iterable]):
    """Find nearest index of asked value.

    Parameters
    ----------
    data : np.array
        input array (e.g. m/z values)
    value : Union[int, float, np.ndarray]
        asked value

    Returns
    -------
    index :
        index value
    """
    data = np.asarray(data)
    if isinstance(value, Iterable):
        return np.asarray([np.argmin(np.abs(data - _value)) for _value in value], dtype=np.int64)
    return np.argmin(np.abs(data - value))


@nb.njit()
def find_nearest_index_array(data: SimpleArrayLike, value: ty.Union[np.ndarray, ty.Iterable]) -> np.ndarray:
    """Find nearest index of asked value.

    Parameters
    ----------
    data : ArrayLike
        input array (e.g. m/z values)
    value : Union[int, float, np.ndarray]
        asked value

    Returns
    -------
    index :
        index value
    """
    data = np.asarray(data)
    return np.asarray([np.argmin(np.abs(data - _value)) for _value in value])


@nb.njit()
def find_nearest_index_single(data: SimpleArrayLike, value: ty.Union[int, float]):
    """Find nearest index of asked value.

    Parameters
    ----------
    data : ArrayLike
        input array (e.g. m/z values)
    value : Union[int, float, np.ndarray]
        asked value

    Returns
    -------
    index :
        index value
    """
    return np.argmin(np.abs(data - value))


def find_nearest_value_single(data: SimpleArrayLike, value: ty.Union[int, float]) -> ty.Union[int, float]:
    """Find nearest value."""
    data = np.asarray(data)
    idx = find_nearest_index_single(data, value)
    return data[idx]


def find_nearest_index_batch(array: SimpleArrayLike, values: SimpleArrayLike) -> np.ndarray:
    """Find nearest index."""
    # make sure array is a numpy array
    array = np.asarray(array)
    values = np.asarray(values)
    if not array.size or not values.size:
        return np.array([])

    # get insert positions
    indices = np.searchsorted(array, values, side="left")

    # find indexes where previous index is closer
    prev_idx_is_less = (indices == len(array)) | (
        np.fabs(values - array[np.maximum(indices - 1, 0)])
        < np.fabs(values - array[np.minimum(indices, len(array) - 1)])
    )
    indices[prev_idx_is_less] -= 1
    return indices


def find_nearest_value(data: ty.Iterable, value: ty.Union[int, float, np.ndarray, Iterable]):
    """Find nearest value."""
    data = np.asarray(data)
    idx = find_nearest_index(data, value)
    return data[idx]


def get_kws(func: ty.Callable, **kwargs) -> ty.Dict:
    """Get kwargs."""
    import inspect

    args = inspect.getfullargspec(func).args

    kws = {}
    for kw in args:
        if kw in kwargs:
            kws[kw] = kwargs[kw]
    return kws


def format_count(value):
    """Format count."""
    if value < 1e3:
        return f"{value:.0f}"
    elif value < 1e6:
        return f"{value/1e3:.1f}K"
    elif value < 1e9:
        return f"{value/1e6:.1f}M"
    return f"{value/1e9:.1f}B"


def format_size(size: int) -> str:
    """Convert bytes to nicer format."""
    if size < 2**10:
        return "%s" % size
    elif size < 2**20:
        return "%.1fK" % (size / float(2**10))
    elif size < 2**30:
        return "%.1fM" % (size / float(2**20))
    elif size < 2**40:
        return "%.1fG" % (size / float(2**30))
    elif size < 2**50:
        return "%.1fT" % (size / float(2**40))
    return "%.1fP" % (size / float(2**50))


def is_number(value):
    """Quick and easy way to check if input is a number.

    Parameters
    ----------
    value : any
        input value

    Returns
    -------
    bool
        returns True/False, depending if a value is a number
    """
    return isinstance(value, (int, float, complex))


def check_value_order(value_min, value_max):
    """Check whether the value order is correct (min -> max).

    Parameters
    ----------
    value_min : int or float or complex
        presumed minimal value
    value_max : int or float or complex
        presumed maximal value

    Returns
    -------
    value_min : int or float or complex
        true minimal value
    value_max : int or float or complex
        true maximal value
    """
    if not is_number(value_min) or not is_number(value_max):
        return value_min, value_max

    if value_max < value_min:
        value_max, value_min = value_min, value_max
    return value_min, value_max


def get_value(new_value, current_value):
    """Get value."""
    if new_value is None:
        return current_value
    return new_value


def rescale(values: ty.Union[np.ndarray, ty.List], new_min: float, new_max: float, dtype=None) -> np.ndarray:
    """Rescale values from one range to another.

    Parameters
    ----------
    values : Union[np.ndarray, List]
        input range
    new_min : float
        new minimum value
    new_max : float
        new maximum value
    dtype :
        data type

    Returns
    -------
    new_values : np.ndarray
        rescaled range
    """
    values = np.asarray(values)
    if dtype is None:
        dtype = values.dtype
    old_min, old_max = get_min_max(values)
    new_values = ((values - old_min) / (old_max - old_min)) * (new_max - new_min) + new_min
    return new_values.astype(dtype)


def rescale_value(
    value: float,
    old_min: float,
    old_max: float,
    new_min: float = 0.0,
    new_max: float = 1.0,
    clip: bool = True,
) -> float:
    """Rescale value to new range.

    Parameters
    ----------
    value : float
        value that needs to be rescaled
    old_min : float
        minimum value of the original range
    old_max : float
        maximum value of the original range
    new_min : float
        minimum value of the new range
    new_max : float
        maximum value of the new range
    clip : bool
        if True, the new value will be clipped to fit inside of the new maximum range
    """
    if clip:
        if value < old_min:
            value = old_min
        if value > old_max:
            value = old_max
    return ((value - old_min) / (old_max - old_min)) * (new_max - new_min) + new_min


def chunks(item_list, n_items: int = 0, n_tasks: int = 0):
    """Yield successive n-sized chunks from `item_list`."""
    if n_items == 0 and n_tasks == 0:
        raise ValueError("You must specified either 'n_items' or 'n_tasks'.")
    if n_tasks:
        n_items = ceil(len(item_list) / n_tasks)
    for i in range(0, len(item_list), n_items):
        yield item_list[i : i + n_items]


def zip_chunks(item_list, n_items: int, *items):
    """Yield successive n-sized chunks from `item_list`."""
    yield from zip(chunks(item_list, n_items), *[chunks(x, n_items) for x in items])


def sequential_chunks(item_list, n_items: int):
    """Create multiple lists of chunks that will be subsequent in nature.

    For instance if the `item_list` has
    the following values: [0, 1, 2, 3] and `n_items=2`, it will generate two lists of [0, 2] and [1, 3]. This can
    be helpful when doing tasks in chunks but want to see results of subsequent chunks.
    """
    n = len(item_list)
    n_pots = ceil(n / n_items)
    pots = [[] for _ in range(n_pots)]
    pot, current = 0, 0
    while current < n:
        pots[pot].append(item_list[current])
        pot += 1
        if pot >= n_pots:
            pot = 0
        current += 1
    yield from pots


def get_min_max(values):
    """Get the minimum and maximum value of an array."""
    return [np.min(values), np.max(values)]


def check_image_orientation(array):
    """Transpose image if the primary size is larger than the secondary size in order to improve images.

    Parameters
    ----------
    array: np.array
        2D heatmap

    Returns
    -------
    zvals: np.array
        (potentially) transposed 2D heatmap
    """
    shape = array.shape
    if len(shape) == 3:
        return np.swapaxes(array, 1, 2) if shape[1] > shape[2] else array
    return array.T if shape[0] > shape[1] else array


def slugify(value, allow_unicode=False):
    """Convert to ASCII if 'allow_unicode' is False.

    Convert spaces or repeated dashes to single dashes. Remove characters that aren't alphanumerics, underscores,
    or hyphens. Convert to lowercase. Also strip leading and trailing whitespace, dashes, and underscores.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    value = re.sub(r"[^.=\w\s-]", "", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")


def get_array_window(array: np.ndarray, min_val: float, max_val: float, *arrays):
    """Get narrower view of array based on upper and lower limits.

    The first array is the one that is used to create mask.
    """
    mask = np.logical_and(array >= min_val, array <= max_val)
    _arrays = [array[mask]]
    for _array in arrays:
        if _array.shape[0] != mask.shape[0]:
            raise ValueError("Incorrect shape of the input arrays")
        _arrays.append(_array[mask])
    return _arrays


def split_array(array: np.ndarray, chunk_size: int, axis=0):
    """Split array according to chunk size."""
    n = array.shape[axis]
    while True:
        if n <= chunk_size:
            yield array
            break
        else:
            if axis == 0:
                yield array.iloc[:chunk_size]
                array = array.iloc[chunk_size:]
            else:
                yield array[:, :chunk_size]
                array = array[:, chunk_size:]
            n -= chunk_size


def _remove_duplicates_from_dict(data):
    """Remove duplicates from list of dictionaries."""
    # list of dictionaries
    if isinstance(data, list):
        # check if list of dictionaries
        if all(isinstance(d, dict) for d in data):
            return [dict(t) for t in {tuple(d.items()) for d in data}]
    return data


class Cycler:
    """Cycling class similar to itertools.cycle with the addition of `previous` functionality."""

    def __init__(self, c):
        self._c = c
        self._index = -1

    def __len__(self):
        return len(self._c)

    def __next__(self):
        self._index += 1
        if self._index >= len(self._c):
            self._index = 0
        return self._c[self._index]

    def next(self):
        """Go forward."""
        return self.__next__()

    def previous(self):
        """Go backwards."""
        self._index -= 1
        if self._index < 0:
            self._index = len(self._c) - 1
        return self._c[self._index]

    def current(self):
        """Get current index."""
        return self._c[self._index]

    def set_current(self, index: int):
        """Set current index."""
        self._index = index



"""Various utility functions speed up using numba."""
import numba as nb
import numpy as np


def nanquantile_along_axis(data: np.ndarray, q1: float, q2: float, axis: int = 0) -> np.ndarray:
    """Numba quantile."""
    return np.vstack(_nanquantile_along_axis(data, q1, q2, axis == 0)).T


@nb.njit(fastmath=True, nogil=True, cache=True)
def _nanquantile_along_axis(data: np.ndarray, q1: float, q2: float, transpose: bool = True):
    """Numba quantile."""
    return [np.nanquantile(d, (q1, q2)) for d in (data.T if transpose else data)]


def quantile_along_axis(data: np.ndarray, q1: float, q2: float, axis: int = 0) -> np.ndarray:
    """Numba quantile."""
    return np.vstack(_quantile_along_axis(data, q1, q2, axis == 0)).T


@nb.njit(fastmath=True, nogil=True, cache=True)
def _quantile_along_axis(data: np.ndarray, q1: float, q2: float, transpose: bool = True):
    """Numba quantile."""
    return [np.quantile(d, (q1, q2)) for d in (data.T if transpose else data)]


def _precompile():
    """Precompile numba functions."""
    import os

    if os.environ.get("KOYO_JIT_PRE", "0") == "0" or os.environ.get("IMIMSPY_JIT_PRE", "0") == "0":
        return
    _nanquantile_along_axis(np.random.rand(100, 10), 0.25, 0.75)
    _quantile_along_axis(np.random.rand(100, 10), 0.25, 0.75)


_precompile()



"""Test secret"""
import numpy as np
import pytest
from koyo.secret import get_short_hash, get_unique_str, hash_iterable, hash_parameters


def test_get_unique_str():
    value = get_unique_str()
    assert isinstance(value, str)


@pytest.mark.parametrize("n", (6, 10))
def test_get_short_hash(n):
    value = get_short_hash(n)
    assert isinstance(value, str)
    assert len(value) == n


def test_hash_iterable():
    assert (
        hash_iterable([0, 1, 2]) == hash_iterable(np.arange(3)) == hash_iterable([2, 1, 0]) == hash_iterable((2, 1, 0))
    )


def test_hash_parameters():
    res = hash_parameters(a=1, b=2, c=3)
    assert isinstance(res, str)



"""Download functions."""
import os.path
import typing as ty
from pathlib import Path
from urllib import request

from loguru import logger
from tqdm.auto import tqdm

from koyo.timer import report_measure_time
from koyo.typing import PathLike


def download_progress(pbar):
    """Wraps tqdm instance.

    Don't forget to close() or __exit__()
    the tqdm instance once you're done with it (easiest using `with` syntax).

    Example
    -------
    >>> with tqdm(...) as p_bar:
    ...     reporthook = download_progress(p_bar)
    ...     request.urlretrieve(reporthook=reporthook)
    """
    last_b = [0]

    def update_to(b=1, bsize=1, tot_size=None):
        """Update progress.

        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tot_size  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tot_size is not None:
            pbar.total = tot_size
        pbar.update((b - last_b[0]) * bsize)
        last_b[0] = b

    return update_to


def download_file(
    link: str,
    filename: ty.Optional[PathLike] = None,
    output_dir: ty.Optional[PathLike] = None,
):
    """Download a file.

    By default, this function download a file to ~/explorer_data.

    Parameters
    ----------
    link : str
        Name of the file to download or url.
    filename : PathLike
        Name of the file to be saved in case of url.
    output_dir : PathLike
        Download file to the path specified.

    Returns
    -------
    path_to_file : Path
        Path to the downloaded file.
    """
    out_path = Path(__file__).parent
    if "http" in link:
        if filename is None:
            filename = os.path.split(link)[1]
            if "?" in filename:
                filename = filename.split("?")[0]
        assert isinstance(filename, str)
        url = link
    else:
        raise ValueError("Expected a link!")

    output_dir = out_path if not isinstance(output_dir, (str, Path)) else output_dir
    path_to_file = Path(output_dir) / filename
    temp_file = path_to_file.parent / (path_to_file.name + ".temp")
    to_download = not path_to_file.exists()

    # download file if needed
    with report_measure_time() as timer:
        if to_download:
            logger.info(f"Downloading {path_to_file}")
            # Check if directory exists else creates it
            if not os.path.exists(output_dir):
                logger.info(f"Folder {output_dir} created")
                os.makedirs(output_dir)
            # Download file
            with tqdm(
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
                desc=f"Downloading {filename}...",
                mininterval=1.0,
            ) as p_bar:
                reporthook = download_progress(p_bar)
                _, _ = request.urlretrieve(url, temp_file, reporthook=reporthook)
            temp_file.rename(path_to_file)
        else:
            logger.info(f"File already downloaded ({path_to_file}). Skipping")
    logger.info(f"Downloaded '{path_to_file}' in {timer()}.")
    return path_to_file



"""Definitions of types."""
import typing as ty
from pathlib import Path

import numpy as np
from scipy import sparse

__all__ = ("ArrayLike", "PathLike", "SparseArray", "CompressedSparseArray")

ArrayLike = ty.Union[ty.List, ty.Tuple, np.ndarray, sparse.csc_matrix, sparse.csr_matrix]
SimpleArrayLike = ty.TypeVar("SimpleArrayLike", ty.List, np.ndarray, ty.Iterable)
CompressedSparseArray = ty.Union[sparse.csc_matrix, sparse.csr_matrix]
SparseArray = ty.Union[sparse.csc_matrix, sparse.csr_matrix, sparse.coo_matrix]
PathLike = ty.Union[str, Path]



"""Decorators."""
import functools
import warnings


def deprecated(func):
    """A decorator which can be used to mark functions as deprecated.

    It will result in a warning being emitted when the function is used.
    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter("always", DeprecationWarning)  # turn off filter
        warnings.warn(
            f"Call to deprecated function {func.__name__}.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter("default", DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    return new_func



from contextlib import contextmanager

import numpy as np


def get_random_seed() -> int:
    """Get random seed."""
    return np.random.randint(0, np.iinfo(np.int32).max - 1, 1)[0]


@contextmanager
def temporary_seed(seed: int):
    """Temporarily set numpy seed."""
    state = np.random.get_state()
    np.random.seed(seed)
    try:
        yield
    finally:
        np.random.set_state(state)



"""Visuals."""
import numba as nb
import numpy as np
from matplotlib.collections import LineCollection


def fig_to_pil(fig):
    """Convert a Matplotlib figure to a PIL Image and return it."""
    import io

    import matplotlib.pyplot as plt
    from PIL import Image

    buf = io.BytesIO()
    fig.savefig(buf, bbox_inches="tight")
    plt.close(fig)
    buf.seek(0)
    img = Image.open(buf)
    return img


def clear_axes(i: int, axs):
    """Clear axes that were not used."""
    if i < len(axs) - 1:
        for _i in range(i + 1, len(axs)):
            axs[_i].axis("off")


def auto_clear_axes(axs):
    """Automatically clear axes. This checks whether there are any lines or collections."""
    if isinstance(axs, np.ndarray):
        axs = axs.flatten()
    for ax in axs:
        if not ax.lines and not ax.collections:
            ax.axis("off")


@nb.njit()
def convert_to_vertical_line_input(x: np.ndarray, y: np.ndarray):
    """Convert two arrays of the same length to list of [x, 0, y, y] vertices used by the matplotlib LineCollection."""
    return [[(x[i], 0), (x[i], y[i])] for i in range(len(x))]


def vertices_to_collection(
    x: np.ndarray,
    y: np.ndarray,
    color: str = "b",
    line_width=3,
    line_style="-",
    alpha: float = 1.0,
) -> LineCollection:
    """Convert list of [x, 0, y, y] vertices to line collection consumed by matplotlib."""
    xy_values = convert_to_vertical_line_input(x, y)
    line_col = LineCollection(
        xy_values,
        colors=[color] * len(x),
        linewidths=line_width,
        linestyles=line_style,
        alpha=alpha,
    )
    return line_col


def compute_divider(value: float) -> float:
    """Compute divider."""
    divider = 1000000000
    value = abs(value)
    if value == 0:
        return 1
    while value == value % divider:
        divider = divider / 1000
    return len(str(int(divider))) - len(str(int(divider)).rstrip("0"))


def y_tick_fmt(x, pos):
    """Y-tick formatter."""

    def _convert_divider_to_str(value, exp_value):
        value = float(value)
        if exp_value in [0, 1, 2]:
            if abs(value) <= 1:
                return f"{value:.2G}"
            elif abs(value) <= 1000:
                if value.is_integer():
                    return f"{value:.0F}"
                return f"{value:.1F}"
        elif exp_value in [3, 4, 5]:
            return f"{value / 1000:.1f}k"
        elif exp_value in [6, 7, 8]:
            return f"{value / 1000000:.1f}M"
        elif exp_value in [9, 10, 11, 12]:
            return f"{value / 1000000000:.1f}B"

    return _convert_divider_to_str(x, compute_divider(x))


def get_intensity_formatter():
    """Simple intensity formatter."""
    from matplotlib.ticker import FuncFormatter

    return FuncFormatter(y_tick_fmt)


def set_intensity_formatter(ax):
    """Set intensity formatter on axes."""
    ax.yaxis.set_major_formatter(get_intensity_formatter())



"""Utilities for spectrum analysis."""
import typing as ty
from bisect import bisect_left, bisect_right

import numba
import numpy as np

from koyo.typing import SimpleArrayLike


@numba.njit(fastmath=True, cache=True)
def ppm_error(x: ty.Union[float, np.ndarray], y: ty.Union[float, np.ndarray]) -> ty.Union[float, np.ndarray]:
    """Calculate ppm error."""
    return ((x - y) / y) * 1e6


@numba.njit(fastmath=True, cache=True)
def get_window_for_ppm(mz: float, ppm: float) -> float:
    """Calculate window size for specified peak at specified ppm."""
    step = mz * 1e-6  # calculate appropriate step size for specified mz value
    peak_x_ppm = mz
    is_subtract = ppm < 0
    ppm = abs(ppm)
    while True:
        if ((peak_x_ppm - mz) / mz) * 1e6 >= ppm:
            break
        peak_x_ppm += step
    value = peak_x_ppm - mz
    return value if not is_subtract else -value


def ppm_diff(a: np.ndarray, axis=-1) -> np.ndarray:
    """Calculate the ppm difference between set of values in array.

    This function is inspired by `np.diff` which very efficiently computes the difference between adjacent points.
    """
    a = np.asarray(a, dtype=np.float)
    nd = a.ndim
    axis = np.core.multiarray.normalize_axis_index(axis, nd)
    slice1 = [slice(None)] * nd
    slice2 = [slice(None)] * nd
    slice1[axis] = slice(1, None)
    slice2[axis] = slice(None, -1)
    slice1 = tuple(slice1)
    slice2 = tuple(slice2)
    return (np.subtract(a[slice1], a[slice2]) / a[slice2]) * 1e6


@numba.njit(cache=True, fastmath=True)
def find_between(data: SimpleArrayLike, min_value: float, max_value: float):
    """Find indices between windows."""
    return np.where(np.logical_and(data >= min_value, data <= max_value))[0]


@numba.njit(cache=True, fastmath=True)
def find_between_tol(data: np.ndarray, value: float, tol: float):
    """Find indices between window and ppm."""
    return find_between(data, value - tol, value + tol)


@numba.njit(cache=True, fastmath=True)
def find_between_ppm(data: np.ndarray, value: float, ppm: float):
    """Find indices between window and ppm."""
    window = get_window_for_ppm(value, abs(ppm))
    return find_between(data, value - window, value + window)


@numba.njit(cache=True, fastmath=True)
def find_between_batch(array: np.ndarray, min_value: np.ndarray, max_value: np.ndarray):
    """Find indices between specified boundaries for many items."""
    min_indices = np.searchsorted(array, min_value, side="left")
    max_indices = np.searchsorted(array, max_value, side="right")

    res = []
    for i in range(len(min_value)):
        _array = array[min_indices[i] : max_indices[i]]
        res.append(min_indices[i] + find_between(_array, min_value[i], max_value[i]))
    return res


def get_peaklist_window_for_ppm(peaklist: np.ndarray, ppm: float) -> ty.List[ty.Tuple[float, float]]:
    """Retrieve peaklist + tolerance."""
    _peaklist = []
    for mz in peaklist:
        _peaklist.append((mz, get_window_for_ppm(mz, ppm)))
    return _peaklist


def get_peaklist_window_for_da(peaklist: np.ndarray, da: float) -> ty.List[ty.Tuple[float, float]]:
    """Retrieve peaklist + tolerance."""
    _peaklist = []
    for mz in peaklist:
        _peaklist.append((mz, da))
    return _peaklist


def get_mzs_for_tol(mzs: np.ndarray, tol: float = None, ppm: float = None):
    """Get min/max values for specified tolerance or ppm."""
    if tol is None and ppm is None or tol == 0 and ppm == 0:
        raise ValueError("Please specify `tol` or `ppm`.")
    elif tol is not None and ppm is not None:
        raise ValueError("Please only specify `tol` or `ppm`.")

    mzs = np.asarray(mzs)
    if tol:
        mzs_min = mzs - tol
        mzs_max = mzs + tol
    else:
        tol = np.asarray([get_window_for_ppm(mz, ppm) for mz in mzs])
        mzs_min = mzs - tol
        mzs_max = mzs + tol
    return mzs_min, mzs_max


def bisect_spectrum(x_spectrum, mz_value, tol: float) -> ty.Tuple[int, int]:
    """Get left/right window of extraction for peak."""
    ix_l, ix_u = (
        bisect_left(x_spectrum, mz_value - tol),
        bisect_right(x_spectrum, mz_value + tol) - 1,
    )
    if ix_l == len(x_spectrum):
        return len(x_spectrum), len(x_spectrum)
    if ix_u < 1:
        return 0, 0
    if ix_u == len(x_spectrum):
        ix_u -= 1
    if x_spectrum[ix_l] < (mz_value - tol):
        ix_l += 1
    if x_spectrum[ix_u] > (mz_value + tol):
        ix_u -= 1
    return ix_l, ix_u



import pytest
from koyo.json import read_json_data, write_json_data


@pytest.mark.parametrize("indent", (0, 4))
@pytest.mark.parametrize("data", ({"A": 1}, [0, 1, 2], ["A", "B", 3, 4]))
def test_read_write(tmpdir_factory, data, indent):
    path = str(tmpdir_factory.mktemp("json") / "test.json")
    write_json_data(path, data, indent=indent)

    _data = read_json_data(path)
    assert _data == data



"""All utility functions that deal with JSON files."""
import os

import numpy as np

from koyo.typing import PathLike
from koyo.utilities import _remove_duplicates_from_dict

try:
    import ujson as json
except ImportError:
    import json


def default(o):
    """Fixes writing of objects containing numpy dtypes.

    Parameters
    ----------
    o : any
        object to be serialized

    Returns
    -------
    int / float
        converted object

    Raises
    ------
    TypeError
    """
    if isinstance(o, (np.int64, np.int32, np.int16)):
        return int(o)
    elif isinstance(o, (np.float64, np.float32, np.float16)):
        return float(o)
    raise TypeError("Could not convert %s of type %s" % o, type(o))


def read_json_data(filepath: PathLike):
    """Read JSON data and metadata.

    Parameters
    ----------
    filepath : PathLike
        path to the JSON file

    Returns
    -------
    loaded json data
    """
    with open(filepath) as f_ptr:
        json_data = json.load(f_ptr)

    return json_data


def write_json_data(filepath: PathLike, obj, indent=4, check_existing=False):
    """Write data to JSON file.

    Parameters
    ----------
    filepath : PathLike
        path to JSON file
    obj : any
        object containing data
    indent : int, optional
        number of spaces to indent by, by default 4
    check_existing : bool, optional
        if True, existing JSON file data will be merged with the new data object
    """
    if not check_existing or not os.path.exists(filepath):
        with open(filepath, "w") as f_ptr:
            json.dump(obj, f_ptr, indent=indent, default=default)
    else:
        with open(filepath, "r+") as f_ptr:
            data = json.load(f_ptr)

            if isinstance(data, list):
                data.extend(obj)
            elif isinstance(data, dict):
                data.update(obj)

            # remove duplicates
            data = _remove_duplicates_from_dict(data)

            f_ptr.seek(0)  # rewind
            json.dump(data, f_ptr, indent=indent, default=default)
            f_ptr.truncate()
    return filepath



import numpy as np
import pytest
from koyo.numpy import nanquantile_along_axis, quantile_along_axis


@pytest.mark.parametrize("axis", (0, 1))
@pytest.mark.parametrize("array", (np.random.rand(100, 10), np.random.rand(10, 100)))
def test_nanquantile_along_axis(array, axis):
    result = nanquantile_along_axis(array, 0.25, 0.75, axis)
    assert result.shape == (2, array.shape[1 - axis])
    q1, q2 = np.nanquantile(array, (0.25, 0.75), axis=axis)
    assert np.allclose(result[0], q1)
    assert np.allclose(result[1], q2)


@pytest.mark.parametrize("axis", (0, 1))
@pytest.mark.parametrize("array", (np.random.rand(100, 10), np.random.rand(10, 100)))
def test_quantile_along_axis(array, axis):
    result = quantile_along_axis(array, 0.25, 0.75, axis)
    assert result.shape == (2, array.shape[1 - axis])
    q1, q2 = np.quantile(array, (0.25, 0.75), axis=axis)
    assert np.allclose(result[0], q1)
    assert np.allclose(result[1], q2)



import typing as ty

_T = ty.TypeVar("_T")
_K = ty.TypeVar("_K")


class MutableSequence(ty.MutableSequence[_T]):
    """List object that holds objects with optional check if it exists."""

    def __init__(self, data: ty.Iterable[_T] = ()):
        self._list: ty.List[_T] = []
        self.extend(data)

    def insert(self, index: int, value: _T) -> None:
        """Insert formula."""
        self._list.insert(index, self._check(value))

    @ty.overload
    def __getitem__(self, i: int) -> _T:
        ...

    @ty.overload
    def __getitem__(self, s: slice) -> ty.MutableSequence[_T]:
        ...

    def __getitem__(self, i):
        return self._list[i]

    @ty.overload
    def __setitem__(self, i: int, o: _T) -> None:
        ...

    @ty.overload
    def __setitem__(self, s: slice, o: ty.Iterable[_T]) -> None:
        ...

    def __setitem__(self, key, value) -> None:
        if isinstance(key, slice):
            if not isinstance(value, ty.Iterable):
                raise TypeError("Can only assign an iterable to slice.")
            self._list[key] = [self._check(v) for v in value]
        else:
            self._list[key] = self._check(value)

    @ty.overload
    def __delitem__(self, i: int) -> None:
        ...

    @ty.overload
    def __delitem__(self, i: slice) -> None:
        ...

    def __delitem__(self, i) -> None:
        del self._list[i]

    def __len__(self) -> int:
        return len(self._list)

    def __repr__(self):
        return repr(self._list)

    def __eq__(self, other: ty.Any):
        return self._list == other

    def __hash__(self) -> int:
        # it's important to add this to allow this object to be hashable
        # given that we've also reimplemented __eq__
        return id(self)

    def _check(self, value: _T) -> _T:
        """Check whether duplicate items exist in the list."""
        raise NotImplementedError("Must implement method")

    def iter(self):
        """Iterator."""
        yield from self._list


class MutableMapping(ty.MutableMapping[_K, _T]):
    """Mutable mapping instance."""

    def __init__(self, data: ty.Dict[_K, _T] = None):
        self._dict: ty.Dict[_K, _T] = {}
        if data is not None:
            self.update(data)

    def __setitem__(self, k: _K, v: _T) -> None:
        self._dict[k] = v

    def __delitem__(self, v: _K) -> None:
        del self._dict[v]

    def __getitem__(self, k: _K) -> _T:
        return self._dict[k]

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self) -> ty.Iterator[_K]:
        yield from self._dict.keys()



import numpy as np
import pytest
from koyo.utilities import (
    check_image_orientation,
    check_value_order,
    find_nearest_index,
    find_nearest_index_batch,
    find_nearest_index_single,
    find_nearest_value_single,
    format_size,
    get_kws,
    get_min_max,
    is_between,
    is_number,
    rescale,
)


def test_find_nearest_index_single():
    array = np.arange(10)
    assert find_nearest_index_single(array, 0) == 0
    assert find_nearest_index_single(array, 0.5) == 0

    assert find_nearest_index(array, 0) == 0
    assert find_nearest_index(array, 0.5) == 0

    assert find_nearest_index_single(array, 9.5) == find_nearest_index(array, 9.5)


def test_find_nearest_index_batch():
    array = np.arange(10)
    indices = find_nearest_index_batch(array, [0, 0.5, 9.5])
    np.all(indices == [0, 0, 9])


def test_find_nearest_value_single():
    array = np.arange(10)
    assert find_nearest_value_single(array, 0) == 0
    assert find_nearest_value_single(array, 0.5) == 0
    assert find_nearest_value_single(array, 10) == 9
    assert find_nearest_value_single(array, 100) == 9


def test_is_number():
    assert is_number(1)
    assert is_number(1.0)
    assert not is_number("1")
    assert not is_number([1])
    assert not is_number((1,))


def test_get_kws():
    def test_func(a=1, b=2):
        pass

    kws = get_kws(test_func, a=2, c=3)
    assert "c" not in kws


def test_check_value_order():
    v1, v2 = 10, 20
    r1, r2 = check_value_order(v1, v2)
    assert r1 == v1
    assert r2 == v2

    v1, v2 = 100, 50
    r1, r2 = check_value_order(v1, v2)
    assert r1 == v2
    assert r2 == v1


def test_is_between():
    assert is_between(0, 0, 1)
    assert is_between(0.5, 0, 1)
    assert is_between(1, 0, 1)
    assert not is_between(1.1, 0, 1)
    assert not is_between(-0.1, 0, 1)


@pytest.mark.parametrize(
    "values, new_min, new_max",
    (
        [np.arange(0, 1000), 0, 1],
        [np.arange(0, 100), 0, 1000],
        [np.exp([0, 1, 2]), 0, 1],
        [1 / np.exp([0, 1, 2]), 0, 1],
    ),
)
def test_rescale(values, new_min, new_max):
    results = rescale(values, new_min, new_max, dtype="float32")
    assert results.min() == new_min
    assert results.max() == new_max


@pytest.mark.parametrize(
    "values, expected",
    (
        [[0, 1, 2, 3, 4], [0, 4]],
        [np.arange(-4123, 1003), [-4123, 1002]],
        [(12, 324, 51, 45), [12, 324]],
    ),
)
def test_get_min_max(values, expected):
    result = get_min_max(values)
    assert len(result) == 2
    assert result[0] == expected[0]
    assert result[1] == expected[1]


@pytest.mark.parametrize("shape", [(2, 10), (10, 2)])
def test_check_image_orientation(shape):
    good_shape = (2, 10)
    zvals = check_image_orientation(np.zeros(shape))
    assert zvals.shape == good_shape


def test_format_size():
    assert "100" == format_size(100)
    assert "1.0K" == format_size(2**10)
    assert "1.0M" == format_size(2**20)
    assert "1.0G" == format_size(2**30)
    assert "1.0T" == format_size(2**40)
    assert "1.0P" == format_size(2**50)
